﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Text.RegularExpressions;

namespace DarwinDotNet.Expressions
{
    /// <summary>
    /// Represents a strongly-typed node within an expression tree. The node may itself be the root
    /// node, a subtree (contained within other expression trees), or a terminal node.
    /// </summary>
    /// <typeparam name="TSymbol">The type of the symbols that can be represented by the expression.
    /// </typeparam>
    [System.Diagnostics.DebuggerDisplay("{ToString(),nq}")]
    public sealed class Expression<TSymbol>
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="Expression{TSymbol}"/> class, specifying
        /// its symbol, the parent expression, and the capacity of children.
        /// </summary>
        /// <param name="symbol">The symbol to represent at the root node of this expression tree.
        /// </param>
        /// <param name="parent">The expression that is the parent of this instance.</param>
        /// <param name="childrenCapacity">The maximum number of children that the expression node
        /// can contain.</param>
        public Expression(TSymbol symbol, Expression<TSymbol> parent, int childrenCapacity)
        {
            this.Parent = parent;
            this.Symbol = symbol;

            if (childrenCapacity != 0)
                this.Children = new List<Expression<TSymbol>>(childrenCapacity);
        }

        /// <summary>
        /// Initialises a new instance of the <see cref="Expression{TSymbol}"/> class.
        /// </summary>
        public Expression()
        {
            this.Parent = null;
            this.Children = new List<Expression<TSymbol>>(1);
        }

        /// <summary>
        /// Gets or sets the symbol that represents this node in the expression tree.
        /// </summary>
        /// <value>A symbol that represents this node within the expression tree.</value>
        public TSymbol Symbol
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the parent of this expression tree, or <see langword="null"/> if the
        /// expression tree is root level.
        /// </summary>
        /// <value>The parent expression of this node in the expression tree.</value>
        public Expression<TSymbol> Parent
        {
            get;
            set;
        }

        /// <summary>
        /// Gets a list of nodes that are children of this expression.
        /// </summary>
        /// <value>A <see cref="List{T}"/> of the child expression nodes, or <see langword="null"/>
        /// if this expression is a terminal node.</value>
        public List<Expression<TSymbol>> Children
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the maximum number of child nodes that this expression node can contain.
        /// </summary>
        /// <value>The maximum number of child nodes that this expression node can contain.</value>
        public int ChildrenCapacity
        {
            get { return this.Children == null ? 0 : this.Children.Capacity; }
        }

        /// <summary>
        /// Gets whether this expression has any child nodes.
        /// </summary>
        /// <value>True if this expression has child nodes, false if it is a termainal node.</value>
        public bool HasChildren
        {
            get { return this.Children != null; }
        }

        /// <summary>
        /// Gets whether this expression is complete. An expression is complete if the
        /// <see cref="List{T}.Count"/> of the <see cref="Children"/> list equals the value of
        /// <see cref="ChildrenCapacity"/>.
        /// </summary>
        /// <value>True if this expression is complete (or <see cref="Children"/> is
        /// <see langword="null"/>), false otherwise.</value>
        public bool IsComplete
        {
            get { return this.Children == null || this.Children.Count == Children.Capacity; }
        }

        ///// <summary>
        ///// When overriden, creates an expression that is a child of this expression node.
        ///// </summary>
        ///// <param name="symbol">The symbol that represents the child node.</param>
        ///// <returns>The created child expression.</returns>
        //public abstract Expression<TSymbol> CreateChild(TSymbol symbol);

        ///// <summary>
        ///// When overriden, gets the maximum number of children that a node of the specified symbol
        ///// can contain.
        ///// </summary>
        ///// <param name="symbol">The symbol for which to get the corresponding capacity of children.
        ///// </param>
        ///// <returns>The maximum number of children that a node of symbol <paramref name="symbol"/>
        ///// can contain.</returns>
        //public abstract int GetChildrenCapacity(TSymbol symbol);

        /// <summary>
        /// Returns a string representation of this expression and its children (if it has any).
        /// </summary>
        /// <returns>A string representation of this expression.</returns>
        public override string ToString()
        {
            return (this.Children == null) ? this.Symbol.ToString() :
                string.Format("{0} -> {1}", this.Symbol, string.Join(", ", this.Children.Select(
                expr => expr.Symbol.ToString()).ToArray()));
        }
    }
}
