﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet.Expressions
{
    /// <inheritdoc/>
    /// <summary>
    /// Represents the genetic information used to encode an <see cref="Expression{TSymbol}"/>.
    /// </summary>
    /// <typeparam name="TSymbol">The type of the symbols that can be represented by the expression.
    /// </typeparam>
    /// <remarks>
    /// <para>
    /// The <see cref="HeadLength"/> of the chromosome corresponds to the length of the section at
    /// the start of the <see cref="Chromosome.Codons"/> array in which only non-terminal codons
    /// may exist. Similarly, the <see cref="TailLength"/> of the chromosome corresponds to the
    /// length of the section at the end of the <see cref="Chromosome.Codons"/> array (starting
    /// immediately after the head section) in which either terminal or non-terminal codons may
    /// exist.
    /// </para>
    /// </remarks>
    public sealed class ExpressionChromosome<TSymbol> : Chromosome<TSymbol>
    {
        /// <inheritdoc/>
        /// <summary>
        /// Initialises a new instance of the <see cref="ExpressionChromosome{TSymbol}"/> class,
        /// specifying the head and tail lengths. The length of the chromosome is the sum of
        /// <paramref name="headLength"/> and <paramref name="tailLength"/>.
        /// </summary>
        /// <param name="headLength">The length of the head section within the
        /// <see cref="Chromosome.Codons"/> array.</param>
        /// <param name="tailLength">The length of the tail section within the
        /// <see cref="Chromosome.Codons"/> array.</param>
        public ExpressionChromosome(int headLength, int tailLength)
            : base()
        {
            this.HeadLength = headLength;
            this.Codons = new object[headLength + tailLength];
        }

        /// <inheritdoc/>
        /// <summary>
        /// Initialises a new instance of the <see cref="ExpressionChromosome{TSymbol}"/> class,
        /// specifying its length.
        /// </summary>
        private ExpressionChromosome(int length)
            : base(length)
        {
        }

        /// <inheritdoc/>
        /// <summary>
        /// Initialises a new instance of the <see cref="ExpressionChromosome{TSymbol}"/> class.
        /// </summary>
        private ExpressionChromosome()
            : base()
        {
        }

        /// <summary>
        /// Gets the length of the tail section within the <see cref="Chromosome.Codons"/> array.
        /// </summary>
        /// <value>The length of the tail section.</value>
        public int TailLength
        {
            get { return Length - HeadLength; }
        }

        /// <summary>
        /// Gets the length of the head section within the <see cref="Chromosome.Codons"/> array.
        /// </summary>
        /// <value>The length of the head section.</value>
        public int HeadLength
        {
            get;
            private set;
        }

        /// <summary>
        /// Generates an expression tree, using the specified function to create each node.
        /// </summary>
        /// <param name="createExpression">A function that is called to create each expression node
        /// within the tree. It takes as parameters the parent expression and the symbol of the nide
        /// to create. It should return the created child node.</param>
        /// <returns>The root node of the generated expression tree.</returns>
        public Expression<TSymbol> GenerateExpressionTree(Func<Expression<TSymbol>, TSymbol, 
            Expression<TSymbol>> createExpression)
        {
            Expression<TSymbol> rootExpression = null;
            var expressions = new Queue<Expression<TSymbol>>(this.HeadLength);
            Expression<TSymbol> curParentExpr = null;
            Expression<TSymbol> curExpr;
            TSymbol curCodon;

            for (int i = 0; i < this.Length; i++)
            {
                curCodon = this[i];
                //curExpr = new Expression<T>(curCodon, curParentExpr, GetChildrenCount<T>(curCodon));
                curExpr = createExpression(curParentExpr, curCodon);
                if (curParentExpr == null)
                {
                    rootExpression = curExpr;
                    curParentExpr = rootExpression;
                    if (rootExpression.IsComplete)
                        break;
                    continue;
                }
                curParentExpr.Children.Add(curExpr);
                expressions.Enqueue(curExpr);

                while (curParentExpr.IsComplete)
                {
                    if (expressions.Count == 0)
                    {
                        i = this.Length;
                        break;
                    }
                    curParentExpr = expressions.Dequeue();
                }
            }

            return rootExpression;
        }

        /// <inheritdoc/>
        public override Chromosome<TSymbol> Clone(bool copyIsOperable)
        {
            var clone = new ExpressionChromosome<TSymbol>();
            CloneBase(clone, copyIsOperable);
            clone.HeadLength = this.HeadLength;
            return clone;
        }

        /// <summary>
        /// Returns a string representation of this chromosome, using the <c>ToString</c> method
        /// as the transformation function.
        /// </summary>
        /// <returns>A string representation of this chromosome.</returns>
        public override string ToString()
        {
            return ToString(item => item.ToString());
        }

        /// <summary>
        /// Returns a string representation of this chromosome, using the specified transformation
        /// function.
        /// </summary>
        /// <param name="transform">A function that transforms a given symbol into a
        /// <see cref="String"/>.</param>
        /// <returns>A string representation of this chromosome.</returns>
        public string ToString(Func<TSymbol, string> transform)
        {
            return string.Join(string.Empty, Array.ConvertAll(this.Codons, 
                item => transform((TSymbol)item)));
        }
    }
}
