﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet
{
    /// <inheritdoc/>
    /// <typeparam name="TCodon">The type of the codons (fundamental building blocks) that compose
    /// the chromosome.</typeparam>
    /// <seealso cref="Chromosome"/>
    public class Chromosome<TCodon> : Chromosome
    {
        /// <inheritdoc/>
        /// <summary>
        /// Initialises a new instance of the <see cref="Chromosome{TCodon}"/> class, specifying its
        /// length.
        /// </summary>
        public Chromosome(int length)
            : base(length)
        {
        }

        /// <summary>
        /// Initialises a new instance of the <see cref="Chromosome{TCodon}"/> class.
        /// </summary>
        public Chromosome()
            : base()
        {
        }

        /// <inheritdoc cref="Chromosome.this"/>
        public new TCodon this[int position]
        {
            get { return (TCodon)Codons[position]; }
            set { Codons[position] = value; }
        }

        /// <inheritdoc cref="Chromosome.Clone"/>
        public new virtual Chromosome<TCodon> Clone(bool copyIsOperable)
        {
            var clone = new Chromosome<TCodon>();
            CloneBase(clone, copyIsOperable);
            return clone;
        }
    }

    /// <summary>
    /// Represents the genetic information used to encode an individual member of a
    /// <see cref="Population"/>.
    /// </summary>
    /// <remarks>
    /// <para>
    /// A chromosome (or genome) specifies the set of parameters that define a possible solution to
    /// a problem by encoding these parameters in a fixed-length array of objects. Before being
    /// evaluated, chromosomes are usually transformed into a more manageable form.
    /// </para>
    /// </remarks>
    /// <seealso cref="Chromosome{TCodon}"/>
    [System.Diagnostics.DebuggerDisplay("Fitness = {Fitness:#0}")]
    public class Chromosome
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="Chromosome"/> class, specifying its
        /// length.
        /// </summary>
        /// <param name="length">The number of codons that the chromosome should contain.</param>
        public Chromosome(int length)
            : this()
        {
            this.Codons = new object[length];
        }

        /// <summary>
        /// Initialises a new instance of the <see cref="Chromosome"/> class.
        /// </summary>
        public Chromosome()
        {
            this.IsOperable = true;
        }

        /// <summary>
        /// Gets or sets the fitness of the chromosome.
        /// </summary>
        /// <value>A value that represents the absolute fitness of the chromosome within the
        /// population.</value>
        public double Fitness
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets whether the chromosome is currently operable (operators can perform on it)
        /// or inoperable (no operators can perform on it).
        /// </summary>
        /// <value><see langword="true"/> if the chromosome is operable, <see langword="false"/>
        /// otherwise.</value>
        /// <remarks>The value of this property does not affect whether selection operators are
        /// performed on it.</remarks>
        public bool IsOperable
        {
            get;
            set;
        }
        
        /// <summary>
        /// Gets the length of the chromosome.
        /// </summary>
        /// <value>The number of codons contained by the chromosome.</value>
        public int Length
        {
            get { return this.Codons.Length; }
        }

        /// <summary>
        /// Gets or sets the codon located at the specified position.
        /// </summary>
        /// <param name="index">The zero-based position of the codon to get or set.</param>
        public object this[int index]
        {
            get { return Codons[index]; }
            set { Codons[index] = value; }
        }

        /// <summary>
        /// Gets or sets the array of codons contained by the chromosome.
        /// </summary>
        /// <value>The array of codons.</value>
        protected internal object[] Codons
        {
            get;
            set;
        }

        /// <summary>
        /// Creates a new chromsome that is a deep copy of the current instance. 
        /// </summary>
        /// <inheritdoc cref="CloneBase" select="param[@name='copyIsOperable']"/>
        /// <returns>The created copy of the current instance.</returns>
        public virtual Chromosome Clone(bool copyIsOperable)
        {
            var clone = new Chromosome();
            CloneBase(clone, copyIsOperable);
            return clone;
        }

        /// <summary>
        /// Copies all data from the current instance to the specified clone chromosome.
        /// </summary>
        /// <param name="clone">The clone object to which to copy all data of the current instance.
        /// </param>
        /// <param name="copyIsOperable"><see langword="true"/> to copy the value of the
        /// <see cref="IsOperable"/> property to the clone, <see langword="false"/> otherwise.
        /// </param>
        protected void CloneBase(Chromosome clone, bool copyIsOperable)
        {
            clone.Codons = (object[])Codons.Clone();
            clone.Fitness = Fitness;
            if (copyIsOperable)
                clone.IsOperable = IsOperable;
        }
    }
}
