﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet
{
    /// <inheritdoc/>
    /// <typeparam name="TCodon">The type of the codons on which the operator can perform.
    /// </typeparam>
    public abstract class Mutator<TCodon> : Mutator
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="Mutator{TCodon}"/> class.
        /// </summary>
        public Mutator()
            : base()
        {
        }

        /// <inheritdoc/>
        public sealed override object Mutate(Chromosome chromosome, int index, object codon)
        {
            return Mutate((Chromosome<TCodon>)chromosome, index, (TCodon)codon);
        }

        /// <inheritdoc cref="Mutator.Mutate"/>
        public abstract TCodon Mutate(Chromosome<TCodon> chromosome, int index, TCodon codon);
    }

    /// <summary>
    /// Provides the abstract base class from which all operators that perform mutations of the
    /// codons of a chromosome must derive.
    /// </summary>
    public abstract class Mutator : Operator
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="Mutator"/> class.
        /// </summary>
        public Mutator()
            : base()
        {
        }

        /// <inheritdoc/>
        /// <summary>
        /// Mutates each chromosome in the specified population.
        /// </summary>
        public sealed override void Perform(Population population)
        {
            var prob = this.ProbabilityFunction(population);
            var chromosomes = population.Chromosomes;
            Chromosome chromosome;

            for (int i = 0; i < chromosomes.Count; i++)
            {
                chromosome = chromosomes[i] as Chromosome;
                if (chromosome != null && chromosome.IsOperable)
                    Perform(chromosome, prob);
            }
        }

        /// <summary>
        /// Mutates the specified chromosome.
        /// </summary>
        /// <param name="chromosome">The chromosome to be mutated.</param>
        /// <param name="probability">The probability with which to mutate each codon.</param>
        /// <remarks>This method should only be overriden if derived class needs to perform
        /// mutation on the chromosome as a whole rather than individual codons.</remarks>
        public virtual void Perform(Chromosome chromosome, double probability)
        {
            var codons = chromosome.Codons;

            for (int i = 0; i < codons.Length; i++)
            {
                if (this.Randomiser.GetDouble() < probability)
                    codons[i] = Mutate(chromosome, i, codons[i]);
            }
        }

        /// <summary>
        /// When overriden, mutates the codon at the specified index the specified chromosome. This
        /// method should perform mutation with certainty.
        /// </summary>
        /// <param name="chromosome">The chromosome to be mutated.</param>
        /// <param name="index">The index of the codon within the <see cref="Chromosome.Codons"/>
        /// array.</param>
        /// <param name="codon">The current value of the codon to mutate.</param>
        /// <returns>The new value of codon. This should usually be different from
        /// <paramref name="codon"/> (the previous value of the codon) but need not always be.
        /// </returns>
        public abstract object Mutate(Chromosome chromosome, int index, object codon);
    }
}
