﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet
{
    /// <summary>
    /// Provides the abstract base class from which all genetic operators must derive.
    /// </summary>
    /// <remarks>
    /// <para>
    /// This class does not implement any functionality, but rather relies on the derived class to
    /// perform some operation on the entire or part of the population.
    /// </para>
    /// <para>
    /// In general, it is recommended to use the <see cref="SingletonOperator"/> or
    /// <see cref="PairOperator"/> class to implement a custom operator where appropiate, as these
    /// classes provide additional functionality specific to operators that manipulate a single or
    /// pair of chromosomes respectively.
    /// </para>
    /// </remarks>
    public abstract class Operator
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="Operator"/> class, specifying the
        /// probability function for the operator being performed.
        /// </summary>
        public Operator(Func<Population, double> probabilityFunction)
            : this()
        {
            this.ProbabilityFunction = probabilityFunction;
        }

        /// <summary>
        /// Initialises a new instance of the <see cref="Operator"/> class.
        /// </summary>
        public Operator()
        {
        }

        /// <summary>
        /// Gets or sets the function that returns the probability with which the operator should
        /// be performed on a given population.
        /// </summary>
        /// <value>A function that returns the probability that the operator is performed for a
        /// given population.</value>
        public Func<Population, double> ProbabilityFunction
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the object that generates random or pseudo-random numbers.
        /// </summary>
        /// <value>An object whose type implements the <see cref="IRandomiser"/> interface.</value>
        public IRandomiser Randomiser
        {
            get;
            set;
        }

        /// <summary>
        /// When overridden in a derived class, performs the operator on the specified population.
        /// </summary>
        /// <param name="population">The population on which to perform the operator.</param>
        /// <remarks>
        /// This method is always called during the creation of a new generation in the evolution
        /// process. It is the duty of the implementer to decide when to actually perform the
        /// operator (though its probability should be determined by
        /// <see cref="ProbabilityFunction"/>) and otherwise perform nothing. This decision may
        /// occur at any level (population-level, chromosome-level, or codon-level, for example).
        /// </remarks>
        public abstract void Perform(Population population);
    }
}
