﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet
{
    /// <summary>
    /// Provides the abstract base class from which all genetic operators that perform on a pair of
    /// chromsomes at a time should derive.
    /// </summary>
    public abstract class PairOperator : Operator
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="PairOperator"/> class, specifying the
        /// probability function for the operator being performed.
        /// </summary>
        public PairOperator(Func<Population, double> probabilityFunction)
            : base(probabilityFunction)
        {
        }

        /// <summary>
        /// Initialises a new instance of the <see cref="PairOperator"/> class.
        /// </summary>
        public PairOperator()
            : base()
        {
        }

        /// <inheritdoc/>
        /// <summary>
        /// Performs the operator on random pairs of chromosomes within the specified population
        /// with probability determined by <see cref="Operator.ProbabilityFunction"/>. Each
        /// chromosome only exists as part of one particular pair.
        /// </summary>
        public sealed override void Perform(Population population)
        {
            var prob = this.ProbabilityFunction(population);
            var chromosomes = population.Chromosomes;
            var chosen = new bool[chromosomes.Count];
            var firstHalfCount = chromosomes.Count / 2;
            var secondHalfCount = chromosomes.Count - firstHalfCount;
            var partners = new int[secondHalfCount];
            int partnerIndex = 0;
            int chromosomeIndex;
            Chromosome chromosome;

            for (int i = 0; i < secondHalfCount; i++)
            {
                chromosomeIndex = firstHalfCount + i;
                chromosome = chromosomes[chromosomeIndex];
                if (chromosome.IsOperable)
                    partners[partnerIndex++] = chromosomeIndex;
                else
                    secondHalfCount--;
            }

            for (int i = 0; i < firstHalfCount; i++)
            {
                chromosome = chromosomes[i];
                if (chromosome.IsOperable && this.Randomiser.GetDouble() < prob)
                {
                    partnerIndex = this.Randomiser.GetInteger(secondHalfCount--);
                    Perform(chromosome, chromosomes[partners[partnerIndex]]);
                    partners[partnerIndex] = partners[secondHalfCount];
                }
            }
        }

        /// <summary>
        /// When overridden in a derived class, performs the operator on the specified pair of 
        /// chromosomes.
        /// </summary>
        /// <param name="chromosomeA">The first chromosome on which to perform the operator.</param>
        /// <param name="chromosomeB">The second chromosome on which to perform the operator.
        /// </param>
        public abstract void Perform(Chromosome chromosomeA, Chromosome chromosomeB);
    }
}
