﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet
{
    /// <summary>
    /// Provides the abstract base class from which all operators that perform recombinations of the
    /// codons of a pair of chromosomes must derive. Recombination involves the interchange of
    /// slices of codons between two chromosomes.
    /// </summary>
    public abstract class Recombinator : PairOperator
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="Recombinator"/> class.
        /// </summary>
        public Recombinator()
            : base()
        {
        }

        /// <inheritdoc/>
        /// <summary>
        /// Recombines random slices of chromosomes between the specified pair of chromosomes.
        /// </summary>
        public sealed override void Perform(Chromosome chromosomeA, Chromosome chromosomeB)
        {
            int index;
            int length;

            Recombine(chromosomeA, chromosomeB, out index, out length);
            InterchangeCodons(chromosomeA.Codons, chromosomeB.Codons, index, length);
        }

        /// <summary>
        /// When overriden, outputs information about the random slices of codons to be interchanged
        /// between the specified pair of chromosomes.
        /// </summary>
        /// <param name="chromosomeA">The first chromosome, from which a slice of codons is to be
        /// copied to <paramref name="chromosomeB"/>.</param>
        /// <param name="chromosomeB">The second chromosome, from which a slice of codons is to be
        /// copied to <paramref name="chromosomeA"/>.</param>
        /// <param name="index">The index (within the <see cref="Chromosome.Codons"/> arrays of both
        /// chromosomes) of the start of the slices of codons to interchange.</param>
        /// <param name="length">The length of the slices of codons to interchange.</param>
        public abstract void Recombine(Chromosome chromosomeA, Chromosome chromosomeB,
            out int index, out int length);

        // Interchanges the specified slices between the two arrays of codons.
        private void InterchangeCodons(object[] codonsA, object[] codonsB, int index, int length)
        {
            var tempSequence = new object[length];
            Array.Copy(codonsA, index, tempSequence, 0, length);
            Array.Copy(codonsB, index, codonsA, index, length);
            Array.Copy(tempSequence, 0, codonsB, index, length);
        }
    }
}
