﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet
{
    /// <summary>
    /// Provides the abstract base class for an operator that perform the selection of a new
    /// generation of chromosomes.
    /// </summary>
    public abstract class Selector : Operator
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="Selector"/> class.
        /// </summary>
        public Selector()
            : base(null)
        {
            this.FitnessComparer = new ChromosomeFitnessComparer();
        }

        /// <summary>
        /// Gets or sets the <see cref="IComparer{Chromosome}"/> that compares chromosomes by their
        /// fitness values.
        /// </summary>
        /// <value>A <see cref="IComparer{Chromosome}"/> that compares chromosomes by their fitness
        /// values.</value>
        protected IComparer<Chromosome> FitnessComparer
        {
            get;
            private set;
        }

        /// <inheritdoc/>
        /// <summary>
        /// Selects the next generation of chromosomes by calling the 
        /// <see cref="SelectNextGeneration"/> method, specifying that they be added to the
        /// <see cref="Population.NewChromosomes"/> list of the specified
        /// <paramref name="population"/>.
        /// </summary>
        /// <remarks>The <see cref="Operator.ProbabilityFunction"/> property is ignored by this
        /// method.</remarks>
        public sealed override void Perform(Population population)
        {
            SelectNextGeneration(population, population.NewChromosomes);
        }

        /// <summary>
        /// When overriden, selects chromosomes to include in the next generation of the specified
        /// population and adds them to the specified list.
        /// </summary>
        /// <param name="population">The population from which to select.</param>
        /// <param name="newChromosomes">A list to which to add chromosomes selected for the new
        /// generation.</param>
        public abstract void SelectNextGeneration(Population population, 
            List<Chromosome> newChromosomes);
    }

    /// <summary>
    /// Compares functionality to compare two chromosomes based on their fitnesses.
    /// </summary>
    public class ChromosomeFitnessComparer : Comparer<Chromosome>
    {
        /// <inheritdoc/>
        /// <summary>
        /// Compares two objects and returns an indication of their relative sort order.
        /// </summary>
        public override int Compare(Chromosome x, Chromosome y)
        {
            if (x == null) return -1;
            if (y == null) return +1;
            return x.Fitness.CompareTo(y.Fitness);
        }
    }
}
