﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet.Selectors
{
    /// <summary>
    /// Represents a selector that chooses a certain number of individuals (elitists) that have the
    /// highest fitnesses within the population and copies them into the subsequent generation.
    /// Elitist chromosomes are not modified in any way when they are copied into the subsequent 
    /// generation.
    /// </summary>
    /// <remarks>
    /// This operator is capable of selecting either a static or dynamic section of the population
    /// as elitists. For a static number, the <see cref="ElitistCount"/> property needs to be set
    /// initially, whereas for a dynamic number, the <see cref="ElitistCountFunction"/> similarly
    /// needs to be set initially. It is generally not recommended that the 
    /// <see cref="ElitistCount"/> property be adjusted during evolution, as 
    /// <see cref="ElitistCountFunction"/> provides a more reliable approach, and is not prone to
    /// multi-threading issues.
    /// </remarks>
    public class ElitistSelector : Selector
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="ElitistSelector"/> class, specifying the
        /// number of elitists.
        /// </summary>
        /// <param name="elitistCount">The constant number of elitists to maintain in the
        /// population.</param>
        public ElitistSelector(int elitistCount)
            : this()
        {
            this.ElitistCount = elitistCount;
        }

        /// <summary>
        /// Initialises a new instance of the <see cref="ElitistSelector"/> class.
        /// </summary>
        public ElitistSelector()
            : base()
        {
        }

        /// <summary>
        /// Gets or sets the number of elitists to maintain in the population.
        /// </summary>
        /// <value>The constant number of elitists.</value>
        public int ElitistCount
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets a function that is used to determine the number of elitists to be copied
        /// into a subsequent generation
        /// </summary>
        /// <value>A function that returns the number of elitists to include in a subsequent
        /// generation for a specified <see cref="Population"/>.</value>
        public Func<Population, int> ElitistCountFunction
        {
            get;
            set;
        }

        /// <inheritdoc/>
        /// <summary>
        /// Selects the elitist chromosomes of the current generation of the specified population,
        /// and adds them to the specified list.
        /// </summary>
        public override void SelectNextGeneration(Population population, 
            List<Chromosome> newChromosomes)
        {
            var chromosomes = population.Chromosomes;
            var count = chromosomes.Count;
            var curElitistCount = (this.ElitistCountFunction == null ? this.ElitistCount :
                this.ElitistCountFunction(population));

            // Create priority queue for chromosomes with highest fitnesses.
            var bestChromosomes = new C5.IntervalHeap<Chromosome>(curElitistCount, 
                this.FitnessComparer);

            var minBestFitness = double.MinValue;
            Chromosome chromosome;
            for (int i = 0; i < count; i++)
            {
                chromosome = chromosomes[i];
                // Check whether chromosome is elitist.
                if (bestChromosomes.Count < curElitistCount || chromosome.Fitness > minBestFitness)
                {
                    // Chromosome is elitist, so insure that it cannot be modified during the
                    // selection process.
                    chromosome.IsOperable = false;
                    bestChromosomes.Add(chromosome);

                    // Remove chromosome with least fitness from priority queue, if queue is full.
                    if (bestChromosomes.Count > curElitistCount)
                    {
                        bestChromosomes.DeleteMin();
                        minBestFitness = bestChromosomes.FindMin().Fitness;
                    }
                    else
                    {
                        if (minBestFitness == double.MinValue || 
                            chromosome.Fitness < minBestFitness)
                            minBestFitness = chromosome.Fitness;
                    }
                }
                else
                {
                    // Chromosome is not elitist, so insure that it is operable.
                    chromosome.IsOperable = true;
                }
            }
            
            foreach (var c in bestChromosomes)
                newChromosomes.Add(c.Clone(true));
        }
    }
}
