﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet.Selectors
{
    /// <summary>
    /// Represents a selector that randomly chooses eacj individual with a probability that is
    /// proportional to its fitness (or a function thereof).
    /// </summary>
    public class FitnessProportionalSelector : Selector
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="FitnessProportionalSelector"/> class.
        /// </summary>
        public FitnessProportionalSelector()
            : base()
        {
        }

        /// <summary>
        /// Gets or sets the function to transform the fitness value of each chromosome, or
        /// <see langword="null"/> to specify no transformation.
        /// </summary>
        /// <value>A function that transforms a fitness value, or <see langword="null"/> if none.
        /// </value>
        public Func<double, double> FitnessTransformFunction
        {
            get;
            set;
        }

        /// <inheritdoc/>
        /// <summary>
        /// Selects chromosomes from the current generation of the specified population, each with
        /// a probability that is proportional to its fitness, and and adds them to the specified
        /// list.
        /// </summary>
        public override void SelectNextGeneration(Population population, 
            List<Chromosome> newChromosomes)
        {
            var chromosomes = population.Chromosomes;
            var count = chromosomes.Count;

            // Calculate the cumulative fitness of each chromosome within the population.
            var cumulativeFitness = new double[count];
            cumulativeFitness[0] = chromosomes[0].Fitness;
            for (int i = 1; i < count; i++)
                cumulativeFitness[i] = cumulativeFitness[i - 1] + 
                    (this.FitnessTransformFunction == null ? chromosomes[i].Fitness : 
                    this.FitnessTransformFunction(chromosomes[i].Fitness));

            // Pick chromosomes and add them to list until it is full.
            var newListCapacity = newChromosomes.Capacity;
            var fitnessSum = cumulativeFitness[count - 1];
            int searchResult;
            while (newChromosomes.Count < newListCapacity)
            {
                // Select random chromosome using probability distribution determined by cumulative
                // fitnesses.
                searchResult = Array.BinarySearch(cumulativeFitness,
                    this.Randomiser.GetDouble() * fitnessSum);
                newChromosomes.Add(chromosomes[searchResult >= 0 ? searchResult : ~searchResult]
                    .Clone(false));
            }
        }
    }
}
