﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet.Selectors
{
    /// <summary>
    /// Represents a selector that randomly chooses each individual with a probability that is
    /// proportional to the inverse of its rank by fitness (or a function thereof) within the entire
    /// population.
    /// </summary>
    public class RankSelector : Selector
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="RankSelector"/> class.
        /// </summary>
        public RankSelector()
            : base()
        {
        }

        /// <summary>
        /// Gets or sets the function to transform the fitness value of each chromosome, or
        /// <see langword="null"/> to specify no transformation.
        /// </summary>
        /// <value>A function that transforms a fitness value, or <see langword="null"/> if none.
        /// </value>
        public Func<double, double> FitnessTransformFunction
        {
            get;
            set;
        }

        /// <inheritdoc/>
        /// <summary>
        /// Selects chromosomes from the current generation of the specified population, each with
        /// a probability that is proportional to the inverse of its rank by fitness, and adds them
        /// to the specified list.
        /// </summary>
        public override void SelectNextGeneration(Population population,
            List<Chromosome> newChromosomes)
        {
            var chromosomes = population.Chromosomes;
            var count = chromosomes.Count;

            // Sort chromosomes by their fitnesses in descending order.
            if (this.FitnessTransformFunction == null)
                chromosomes.Sort(this.FitnessComparer);
            else
                chromosomes.Sort((x, y) => this.FitnessTransformFunction(x.Fitness).CompareTo(
                    this.FitnessTransformFunction(y.Fitness)));

            var firstPlusLastTerm = 1 + count;
            var rankTotalSum = count * firstPlusLastTerm / 2;

            // Pick chromosomes and add them to list until it is full.
            var newListCapacity = newChromosomes.Capacity;

            int cumulativeRank;
            Chromosome chromosome;
            while (newChromosomes.Count < newListCapacity)
            {
                // Select random chromosome by generating random partial sum and finding
                // corresponding index within set of all chromosomes.
                cumulativeRank = this.Randomiser.GetInteger(rankTotalSum);
                chromosome = chromosomes[(2 * cumulativeRank) / firstPlusLastTerm];
                newChromosomes.Add(chromosome.Clone(false));
            }
        }
    }
}
