﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet.Selectors
{
    /// <summary>
    /// Represents a selector that randomly takes a group of individuals of a certain size and
    /// attempts to choose each individual with a certain probability, in descending of fitness.
    /// </summary>
    public class TournamentSelector : Selector
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="TournamentSelector"/> class, specifying the
        /// tournament size.
        /// </summary>
        /// <param name="tournamentSize">The number of chromosomes to participate in each
        /// tournament.</param>
        public TournamentSelector(int tournamentSize)
            : this(tournamentSize, 1d, true)
        {
        }
        
        /// <summary>
        /// TODO
        /// </summary>
        /// <param name="tournamentSize">TODO</param>
        /// <param name="chooseProbability">TODO</param>
        /// <param name="performReplacement">TODO</param>
        public TournamentSelector(int tournamentSize, double chooseProbability, 
            bool performReplacement)
            : base()
        {
            this.TournamentSize = tournamentSize;
            this.ChooseProbability = chooseProbability;
            this.PerformReplacement = performReplacement;
        }

        /// <summary>
        /// TODO
        /// </summary>
        /// <value>TODO</value>
        public int TournamentSize
        {
            get;
            set;
        }

        /// <summary>
        /// TODO
        /// </summary>
        /// <value>TODO</value>
        public double ChooseProbability
        {
            get;
            set;
        }

        /// <summary>
        /// TODO
        /// </summary>
        /// <value>TODO</value>
        public bool PerformReplacement
        {
            get;
            set;
        }

        /// <inheritdoc/>
        /// <summary>
        /// TODO
        /// </summary>
        public override void SelectNextGeneration(Population population, 
            List<Chromosome> newChromosomes)
        {
            var chromosomes = population.Chromosomes;
            var count = chromosomes.Count;

            int selectionMaxIndex = 0; // exclusive maximum
            int[] selectionIndices = null;
            if (!this.PerformReplacement)
                selectionIndices = new int[count];

            // Create priority queue for chromosomes with highest fitnesses.
            var participants = new C5.IntervalHeap<Chromosome>(this.TournamentSize,
                this.FitnessComparer);

            // Pick chromosomes and add them to list until it is full.
            var newListCapacity = newChromosomes.Capacity;
            int curSelectionIndex;
            while (newChromosomes.Count < newListCapacity)
            {
                if (!this.PerformReplacement)
                {
                    // Recreate array that stores selected chromosomes, thus allowing ones already
                    // chosen possibly to be chosen again.
                    selectionMaxIndex = count;
                    for (int i = 0; i < selectionMaxIndex; i++)
                        selectionIndices[i] = i;
                }

                // Pick a fixed number of chromosomes to participate in the current tournament.
                for (int i = 0; i < this.TournamentSize; i++)
                {
                    if (this.PerformReplacement)
                    {
                        // Add chromosome to tournament, allowing it to be chosen again.
                        participants.Add(chromosomes[this.Randomiser.GetInteger(count)]);
                    }
                    else
                    {
                        // Add chromsome to tournament, disallowing it to be chosen again
                        // (unless array of selections needs to be recreated).
                        curSelectionIndex = this.Randomiser.GetInteger(selectionMaxIndex);
                        participants.Add(chromosomes[selectionIndices[curSelectionIndex]]);
                        selectionIndices[curSelectionIndex] = selectionIndices[--selectionMaxIndex];
                    }
                }

                // Attempt to choose each chromosome within tournament in turn.
                // If none are chosen, ignore the current tournament and create another one.
                while (participants.Count > 0)
                {
                    if (this.Randomiser.GetDouble() < this.ChooseProbability)
                    {
                        newChromosomes.Add(participants.DeleteMax().Clone(false));
                        break;
                    }
                }
            }
        }
    }
}
