﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet
{
    /// <summary>
    /// Provides the abstract base class from which all genetic operators that perform on a single
    /// chromsome at a time should derive.
    /// </summary>
    public abstract class SingletonOperator : Operator
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="SingletonOperator"/> class, specifying
        /// the probability function for the operator being performed.
        /// </summary>
        public SingletonOperator(Func<Population, double> probabilityFunction)
            : base(probabilityFunction)
        {
        }

        /// <summary>
        /// Initialises a new instance of the <see cref="SingletonOperator"/> class.
        /// </summary>
        public SingletonOperator()
            : base()
        {
        }

        /// <inheritdoc/>
        /// <summary>
        /// Performs the operator on each operable chromosome within the specified population
        /// with probability determined by <see cref="Operator.ProbabilityFunction"/>.
        /// </summary>
        public sealed override void Perform(Population population)
        {
            var prob = this.ProbabilityFunction(population);
            var chromosomes = population.Chromosomes;
            Chromosome chromosome;

            for (int i = 0; i < chromosomes.Count; i++)
            {
                chromosome = chromosomes[i];
                if (chromosome.IsOperable && this.Randomiser.GetDouble() < prob)
                    Perform(chromosome);
            }
        }

        /// <summary>
        /// When overridden in a derived class, performs the operator on the specified chromosome.
        /// </summary>
        /// <param name="chromosome">The chromosome on which to perform the operator.</param>
        public abstract void Perform(Chromosome chromosome);
    }
}
