﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet
{
    /// <summary>
    /// Provides the abstract base class from which all operators that perform transpositions of the
    /// codons of a chromosome must derive. Transposition involves the movement of slices of codons
    /// from one position to another.
    /// </summary>
    public abstract class Transpositor : SingletonOperator
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="Transpositor"/> class.
        /// </summary>
        public Transpositor()
            : base()
        {
        }

        /// <inheritdoc/>
        /// <summary>
        /// Transposes a random slice of codons within the specified chromosome.
        /// </summary>
        public sealed override void Perform(Chromosome chromosome)
        {
            int sourceIndex;
            int destinationIndex;
            int length;
            bool insert;

            Transpose(chromosome, out sourceIndex, out destinationIndex, out length, out insert);
            MoveCodons(chromosome.Codons, sourceIndex, destinationIndex, length, insert);
        }

        /// <summary>
        /// When overriden, outputs information about a random slice of codons to be transposed
        /// within the specified chromosome.
        /// </summary>
        /// <param name="chromosome">The chromosome containing the codons to transpose.</param>
        /// <param name="sourceIndex">The index (within the <see cref="Chromosome.Codons"/> array of
        /// the chromosome) of the start of the slice of codons to transpose.</param>
        /// <param name="destinationIndex">The index within the <see cref="Chromosome.Codons"/>
        /// array from which to insert the slice of codons.</param>
        /// <param name="length">The length of the slice of codons to transpose.</param>
        /// <param name="insert"><see langword="true"/> to insert the slice of codons without
        /// overwriting existing codons (thus expanding the size of the chromosome),
        /// <see langword="false"/> to overwrite existing codons (thus maintaining the size of the
        /// chromosome).</param>
        public abstract void Transpose(Chromosome chromosome, out int sourceIndex, 
            out int destinationIndex, out int length, out bool insert);

        // Moves the specified slice within specified array of codons to a new position.
        private void MoveCodons(object[] codons, int sourceIndex, int destinationIndex, int length, 
            bool insert)
        {
            if (insert)
            {
                // Insert given slice of codons without removing any other codons.
                var tempSequence = new object[length];
                Array.Copy(codons, sourceIndex, tempSequence, 0, length);
                Array.Copy(codons, sourceIndex, codons, sourceIndex + length, 
                    codons.Length - sourceIndex - length);
                Array.Copy(tempSequence, 0, codons, destinationIndex, length);
            }
            else
            {
                // Overwrite existing codons with given slice of codons.
                Array.Copy(codons, sourceIndex, codons, destinationIndex, length);
            }
        }
    }
}
