using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;
using Box2DX;
using Box2DX.Collision;
using Box2DX.Common;
using Box2DX.Dynamics;
using Xna = Microsoft.Xna.Framework;
using XnaGraphics = Microsoft.Xna.Framework.Graphics;

namespace CatapultSimulation
{
    public class CatapultGame : Game
    {
        World world;
        BasicEffect basicEffect;

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        public CatapultGame()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            var worldAABB = new AABB();
            worldAABB.LowerBound.Set(-200.0f, -100.0f);
            worldAABB.UpperBound.Set(200.0f, 200.0f);
            Vec2 gravity = new Vec2();
            gravity.Set(0.0f, -10.0f);
            bool doSleep = true;
            world = new World(worldAABB, gravity, doSleep);

            Body ground = null;
            {
                BodyDef bd = new BodyDef();
                bd.Position.Set(0.0f, -10.0f);
                ground = world.CreateBody(bd);

                PolygonDef sd = new PolygonDef();
                sd.SetAsBox(50.0f, 10.0f);
                ground.CreateShape(sd);
            }

            {
                PolygonDef sd = new PolygonDef();
                sd.SetAsBox(0.6f, 0.125f);
                sd.Density = 20.0f;
                sd.Friction = 0.2f;

                RevoluteJointDef jd = new RevoluteJointDef();
                jd.CollideConnected = false;

                const float y = 25.0f;
                Body prevBody = ground;
                for (int i = 0; i < 30; ++i)
                {
                    BodyDef bd = new BodyDef();
                    bd.Position.Set(0.5f + i, y);
                    Body body = world.CreateBody(bd);
                    body.CreateShape(sd);
                    body.SetMassFromShapes();

                    Vec2 anchor = new Vec2(i, y);
                    jd.Initialize(prevBody, body, anchor);
                    world.CreateJoint(jd);

                    prevBody = body;
                }
            }

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            basicEffect = new BasicEffect(GraphicsDevice, null);
            basicEffect.VertexColorEnabled = true;
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            basicEffect.Dispose();
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            world.Step((float)gameTime.ElapsedRealTime.TotalSeconds, 8, 1);
            world.Validate();

            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(XnaGraphics.Color.CornflowerBlue);

            GraphicsDevice.VertexDeclaration = new VertexDeclaration(GraphicsDevice,
                VertexPositionColor.VertexElements);
            GraphicsDevice.RenderState.CullMode = CullMode.None;

            basicEffect.Projection = Matrix.CreateOrthographicOffCenter(0, GraphicsDevice.Viewport.Width,
                GraphicsDevice.Viewport.Height, 0, 0, 1);
            basicEffect.View = Matrix.Identity;
            basicEffect.Begin();
            foreach (var pass in basicEffect.CurrentTechnique.Passes)
            {
                pass.Begin();

                foreach (var bodies in world.GetBodies())
                {
                    foreach (var shape in bodies.GetShapes())
                    {
                        switch (shape.GetType())
                        {
                            case ShapeType.CircleShape:
                                var circleShape = (CircleShape)shape;

                                break;
                            case ShapeType.PolygonShape:
                                var polygonShape = (PolygonShape)shape;
                                
                                var vertices = polygonShape.GetVertices().Take(polygonShape.VertexCount)
                                    .Select(v => new Vector2(v.X * 100 + 100, v.Y * 100 + 100)).ToArray();
                                this.GraphicsDevice.DrawUserPrimitives(PrimitiveType.LineStrip, vertices, 0,
                                    vertices.Length - 1);

                                break;
                        }
                    }
                }

                pass.End();
            }
            basicEffect.End();

            base.Draw(gameTime);
        }
    }
}
