﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using DarwinDotNet;
using DarwinDotNet.Expressions;
using DarwinDotNet.Selectors;

namespace SymbolicRegression
{
    internal static class Program
    {
        private static readonly double maxFitness = 500000;
        private static readonly Func<double, double, double, double> testFunction =
            (x, y, z) => x * x * z - x * y * z + x * y * y - z;
        private static TestCase[] testCases;

        private static IRandomiser randomiser = new BasicRandomiser();

        private static void Main(string[] args)
        {
            // Create test cases for fitness evaluation.
            TestCase curCase;
            testCases = new TestCase[30];

            for (int i = 0; i < testCases.Length; i++)
            {
                curCase = new TestCase()
                {
                    X = randomiser.GetDouble() * 100,
                    Y = randomiser.GetDouble() * 100,
                    Z = randomiser.GetDouble() * 100
                };
                curCase.ExpectedResult = testFunction(curCase.X, curCase.Y, curCase.Z);
                testCases[i] = curCase;
            }

            // Create initial population and then evolve it.
            var population = new Population(2000);

            population.GeneticOperators = new Operator[] { 
                new ElitistSelector() { ElitistCountFunction = 
                    (p => p.GenerationNumber < 80 ? 500 : 250) },
                new FitnessProportionalSelector(),
                new SimpleExpressionMutator() { ProbabilityFunction = (p => 0.14d) },
                new SimpleExpressionTranspositor() { ProbabilityFunction = (p => 0.25d) },
                new SimpleExpressionRecombinator() { ProbabilityFunction = 
                    (p => p.GenerationNumber < 120 ? 0.92 : 0.0)},
            };
            population.SetRandomiser(new BasicRandomiser());
            population.Generate(GenerateRandomChromosome);
            population.Evolve(RunSimulation, maxFitness - 0.5d);

            // Print out results.
            population.Chromosomes.Sort(new ChromosomeFitnessComparer());
            var bestChromosome = population.Chromosomes.Last();
            Console.WriteLine("Found solution in generation {0:#0}.", population.GenerationNumber);
            Console.WriteLine("Best chromosome in final generation: {0} (fitness {1:#0})",
                bestChromosome, bestChromosome.Fitness);

#if DEBUG
            Console.ReadKey(true);
#endif
        }

        public static void RunSimulation(Population population)
        {
            var chromosomes = population.Chromosomes;
            Func<Expression<MathSymbol>, MathSymbol, Expression<MathSymbol>>
                generateExpression = (parent, symbol) => new Expression<MathSymbol>(symbol, parent,
                    GetChildrenCapacity(symbol));

            //for (int i = 0; i < chromosomes.Count; i++)
            Parallel.For(0, chromosomes.Count, delegate(int i)
            {
                var chromosome = (ExpressionChromosome<MathSymbol>)chromosomes[i];
                var expression = chromosome.GenerateExpressionTree(generateExpression);
                TestCase curCase;
                double sumResidualsSquared = 0;

                for (int j = 0; j < testCases.Length; j++)
                {
                    curCase = testCases[j];
                    sumResidualsSquared += (expression.Evaluate(curCase.X, curCase.Y, curCase.Z) -
                        curCase.ExpectedResult).Square();
                }

                if (!double.IsNaN(sumResidualsSquared))
                {
                    chromosome.Fitness = Math.Max(maxFitness - Math.Sqrt(sumResidualsSquared /
                        testCases.Length), 0);
                }
            });

            Console.WriteLine("Generation #{0:#00000} | Best Fitness: {1:#00000} | " + 
                "Average Fitness: {2:#000000}",
                population.GenerationNumber, population.Chromosomes.Max(c => c.Fitness),
                population.Chromosomes.Average(c => c.Fitness));
        }

        public static Chromosome GenerateRandomChromosome(Population population)
        {
            var chromosome = new ExpressionChromosome<MathSymbol>(10, 11);
            for (int i = 0; i < chromosome.HeadLength; i++)
                chromosome[i] = (MathSymbol)randomiser.GetInteger(0, (int)MathSymbol.Max);
            for (int i = chromosome.HeadLength; i < chromosome.Length; i++)
                chromosome[i] = (MathSymbol)randomiser.GetInteger((int)MathSymbol.FunctionMax, 
                    (int)MathSymbol.Max);

            return chromosome;
        }

        private static int GetChildrenCapacity(MathSymbol symbol)
        {
            switch (symbol)
            {
                case MathSymbol.Add:
                    return 2;
                case MathSymbol.Subtract:
                    return 2;
                case MathSymbol.Multiply:
                    return 2;
                case MathSymbol.Divide:
                    return 2;
            }

            return 0;
        }
    }

    public struct TestCase
    {
        public double ExpectedResult;
        public double X, Y, Z;
    }
}
